%**********************************************************
%   Tree View and navigation for disk files
%**********************************************************

%---    Required Prototypes  ------------------------------

define tree_branch (ddir, instr)    ;
define tree_count_subs (ddir)       ;
define tree_create_menu ()          ;
define tree_descend ()              ;
define tree_keymap ()               ;
define tree_node_path ()            ;
define tree_parent_show ()          ;
define tree_point2 ()               ;
define tree_quit ()                 ;
define tree_refresh ()              ;
define tree_show (root)             ;
define tree_sibling_prev ()         ;
define tree_to_listing ()           ;

%----------------------------------------------------------

%   NOTE
%   goto_column_best_try is used because it works
%   even if the buffer is read-only.

%   _traceback  = 1;

%----------------------------------------------------------
%   Variables
%----------------------------------------------------------

%   variable TreeView is defined in site.sl

private
variable tree_cwd           = Null_String;
    %   current directory

variable
    m_s     = "- ",         %   minus space
    p_s     = "+ ",         %   plus  space
    s_s     = "  ",         %   space space
    vline   = char(124),    %   segment | for vertical line
    cornr   = char(96),     %   backtick ` for lower right corner
    vline_m = vline + m_s,  %   child, read
    vline_p = vline + p_s,  %   child, unread
    vline_s = vline + s_s,  %   vline and space
    cornr_m = cornr + m_s,  %   last child, read
    cornr_p = cornr + p_s;  %   last child, unread

%----------------------------------------------------------
%   Functions
%----------------------------------------------------------

define dtr ()
%----------------------------------------------------------
%   Tree view of a directory with first level subdirectories
%   Entry point from the outside.
%----------------------------------------------------------
{
    %   char(0x2502)    full vertical stroke
    %   char(0x2514)    lower right corner
    %   char(0x2518)    lower left corner
    %                   2518 2524 2551 00bb
    %   char(124)       pipe character |
    %   char(96)        grave accent `

    if(_slang_utf8_ok)
    {
        vline_m  = char(0x2502) + m_s;
        vline_p  = char(0x2502) + p_s;
        vline_s  = char(0x2502) + s_s;
        cornr_m  = char(0x2551) + m_s;
        cornr_p  = char(0x2551) + p_s;

        %   2514 is the correct one for corners but it
        %   will merge with a minus following it
        %   let's take one which is not in the font
        %   and delivers the not-found character
    }

    else
    {
        vline_m  = char(124) + m_s;
        vline_p  = char(124) + p_s;
        vline_s  = char(124) + s_s;
        cornr_m  = char(96)  + m_s;
        cornr_p  = char(96)  + p_s;
    }

    %   save current dir, position tree on it

    variable dir=Null_String;
    (,dir,,) = getbuf_info();
    if (dir==Null_String) tree_cwd = getcwd();
    else tree_cwd = dir;

    %   tree may have been requested from dls

    %   tree_cwd does have a terminating slash

    %   close file listing, if applicable
    %   delbuf modified, does not ask questions
    delbuf (DirListing);

    %   prepare for the tree view
    sw2buf (TreeView);
    set_buffer_no_autosave ();
    tree_keymap();
    tree_create_menu();
    hide_status_indicators(1);

    %   display view of current directory
    tree_show (tree_cwd);

    %   when launching treeview from edit buffer
    %   and dir has no subdirs
    ifnot (tree_count_subs (tree_cwd))
    {
        set_readonly (0);
        bob();
        eol();
        newline();
        newline();
        insert("    No sub-dirs, press any key.");
        bob();
        set_readonly (1);
        update(1);
        flush_input(getkey);
        tree_parent_show();
    }
}


define tree_test()
{
}

define tree_pull_menus (keypress)
{
    if (keypress=="f")
        menu_select_menu ("TreeView.&File");

    else if (keypress=="d")
        menu_select_menu ("TreeView.&Display");

    else if (keypress=="h")
        menu_select_menu ("TreeView.&Help");
}

define tree_goto_name()
{
    %   point to initial of text in tree line
    %   it must go one down first

    ifnot(down(1)) return -1;
    return (-2 + tree_point2);
}

define tree_jump_to_initial ()
%----------------------------------------------------------
%   Pressing a letter takes the cursor to the next line
%   with directory beginning with that letter.
%   Upper/lower case makes no difference.
%----------------------------------------------------------
{
    %   catch alt-<letter> for the pulldowns

    variable chk =
        LASTKEY=="f" or LASTKEY=="d" or LASTKEY=="h";

    if(chk)
    {
        variable s = substr(LASTKEY,2,1);
        if(strlen(s)) tree_pull_menus (s);
        return;
    }

    if(LASTKEY[0]==27 and strlen(LASTKEY)==2) return;

    %------------------------------------------------------

    current_mini_cmd = _function_name;

    variable
        initl     = strlow(LASTKEY),
        s_initial = Null_String;

    push_mark();
    forever
    {
        if(tree_goto_name < 0)
        {
            pop_mark(1);
            flush("Not found.");
            return;
        }
        s_initial = strlow(char(what_char()));
        if (s_initial == initl) break;
    }
    pop_mark(0);
}

define tree_menubar()
{
    select_pulldown_file();
}

define tree_help()
{
    select_pulldown_help();
}

define tree_display_help()
{
    variable
        stringA="TreeView",
        stringB="press F1",
        stringC="&Tree View";

    display_help (stringA,stringB,stringC);
    dtr ();
}

define tree_save()
{
    current_mini_cmd = _function_name;
    variable b_name = Null_String;
    variable L;

    b_name = read_path_from_mini ("Save to file:");

    %   aborted, last press was not Return
    if (LAST_CHAR != 13) return;

    %   if response empty, do nothing
    L = strlen(b_name);
    ifnot(L) return;
    if(b_name[L-1] == '/') return;

    %   select entire listing for saving

    push_spot();
    bob();
    push_mark();
    eob();

    %   check file status before proceeding

    variable b_status = file_status (b_name);

    if (b_status == 2)
        error ("Name already in use for a directory");

    else if (b_status == -1)
        error ("File exists and is not accessible");

    else if (b_status == -2)
        error ("Invalid path");

    else if (b_status == -3)
        error ("Undetermined error");

    else if (b_status == 0)
    {
        write_region_to_file (b_name);
    }

    else if (b_status == 1)
    {
        variable ch = get_mini_response
        ("File exists: append/overwrite/cancel? (a/o/c) ");

        if (ch=='a' or ch=='A')
            append_region_to_file (b_name);

        else if (ch=='o' or ch=='O')
            write_region_to_file (b_name);
    }
    pop_spot();
    tree_refresh();
}

define tree_to_listing()
{
    if(ALTKEY_PRESSED)
    {
        tree_quit();
        return;
    }

   variable s = tree_node_path();
   delbuf (whatbuf);
   dls (s);
}

define tree_point2()
%----------------------------------------------------------
%   Go to first character of current name in tree
%----------------------------------------------------------
{
    bol();

    if
    (
        ffind (vline_p) or
        ffind (vline_m) or
        ffind (cornr_p) or
        ffind (cornr_m)
    )
        go_right(3);
        %   move to name

    return what_column();

    %   return of 1 means nothing found
}


define tree_descend()
%----------------------------------------------------------
%   If current node has any children displayed,
%   move to last displayed child.
%   It works even if there is no next sibling at bottom.
%   Only used when collapsing.
%----------------------------------------------------------
{
    while(is_whiteline) go_up(1);

    variable cn = tree_point2();

    forever
    {
        ifnot(down(1)) break;
        if(tree_point2() <= cn)
        {
            go_up(1);
            break;
        }
    }
}

define tree_pgup ()
{
    variable start_line = window_line ();
    %   cursor is on start_line

    goto_top_of_window ();
    %   move to top of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        %   already top, now paginate
        call ("page_up");
        goto_top_of_window ();
    }
}

define tree_pgdn ()
{
    variable start_line = window_line ();
    %   cursor is on start_line

    goto_bottom_of_window ();
    %   move to bottom of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        % already at bottom of window, now paginate
        call ("page_down");
        goto_bottom_of_window ();
    }
}

define tree_sibling_next()
%----------------------------------------------------------
%   Shift not pressed:  to next sibling in tree
%   Shift pressed:      to previous sibling in tree
%----------------------------------------------------------
{
    if (SHIFTKEY_PRESSED)
    {
        return tree_sibling_prev();
    }

    %   Go to next sibling
    %   return is positive if successful, 0 otherwise.

    if(what_line()==1) return 0;

    %   line is a node, not top line

    variable col0, col2, r=1;

    col0 = tree_point2();

    push_mark();

    %   go down indefinitely
    %   at each step compare tree_point2 with col0

    while (down(1))
    {
        col2 = tree_point2();

        if (col0 < col2) continue;

        else if (col0 > col2)
        {
            r=0;
            break;
        }

        else break;
    }

    pop_mark (1-r);

    if(r) r = col2;

    return r;
}

define tree_sibling_prev()
%----------------------------------------------------------
%   To previous sibling in tree
%----------------------------------------------------------
{
    if(what_line()==1) return 0;

    variable col0, col2, r=1;

    col0 = tree_point2();
    push_mark();

    %   go up indefinitely
    %   at each step compare tree_point2 with col0

    while (up(1))
    {
        col2 = tree_point2();

        if (col0 < col2) continue;

        else if (col0 > col2)
        {
            r=0;
            break;
        }

        else break;
    }

    pop_mark (1-r);

    if(r) r = col2;

    return r;
}

define tree_move_down()
{
    go_down(1);
    if(is_whiteline) go_up(1);
    () = tree_point2();
}

define tree_move_up()
{
    go_up(1);
    () = tree_point2();
}

define tree_node_show()
%----------------------------------------------------------
%   Anchor tree at current node
%   provided there is something to list
%----------------------------------------------------------
{
    if (what_line()==1) return;

    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    variable s = tree_node_path();

    ifnot(tree_count_subs (s))
    {
        %   nothing to list
        flush ("No subdirectories.");

        () = tree_point2();
        call ("previous_char_cmd");
        call ("previous_char_cmd");
        del();
        insert ("-");
        () = tree_point2();
        return;
    }

    %   if control here, s has something to list

    tree_show (s);
}

define tree_node_show_path()
%----------------------------------------------------------
%   Will flush the absolute path name of current tree node.
%----------------------------------------------------------
{
   flush (tree_node_path());
}

define tree_parent()
%----------------------------------------------------------
%    Go to the parent of current node in tree
%----------------------------------------------------------
{
    if(what_line == 1 ) return;

    variable cn = tree_point2();

    forever
    {
        if(tree_point2() < cn) break;
        go_up(1);
    }
}

define tree_parent_show()
%----------------------------------------------------------
%   Will anchor tree at parent of current root.
%   Used for tree_upper_dir and tree_collapse
%   when cursor on top line.
%----------------------------------------------------------
{
    variable root, offset, tmp_node;
    tmp_node = tree_node_path();

    if (tree_cwd != "/")
    {
        offset = string_match (tree_cwd, "[/][^/]*[/]$", 1);
        root = substr (tree_cwd, 1, offset);
        tree_show (root);

        % highlight sub-directory we are targeting
        tmp_node = p_s + substr(tmp_node, offset+1,-1);
        ifnot(fsearch (tmp_node)) bob();
    }
}

define tree_upper_dir()
%----------------------------------------------------------
%   To upper directory without collapsing.
%   If on top line, display branch one up.
%----------------------------------------------------------
{
    while (is_whiteline) go_up(1);

    if (what_line() == 1) tree_parent_show();
    else tree_parent();
}

define tree_quit()
{
    sw2buf ("*scratch*");
    delbuf (TreeView);
}

define tree_refresh()
{
   tree_show (tree_cwd);
}

define tree_top_root()
{
    tree_show ("/");
}

define tree_user_root()
{
    variable homedir = dircat(getenv ("HOME"), Null_String);
    tree_show (homedir);
}

define tree_node_name()
%----------------------------------------------------------
%    Returns the name of the current node in tree
%----------------------------------------------------------
{
    variable
        retstr = Null_String,
        ll     = 0;

    if (what_line() == 1)
    {
        ll = strlen (tree_cwd);
        if (ll==1) return "/";
        else retstr=substr (tree_cwd, 1, ll-1);
        %   ll is positive
    }

    () = tree_point2();
    push_mark();
    eol();
    retstr=bufsubstr();

    return retstr;
}

define tree_node_path()
%----------------------------------------------------------
%   Returns the complete pathname of current node.
%----------------------------------------------------------
{
    variable path = Null_String;

    path = tree_node_name();

    if(substr(path,1,1)=="/") return path;
    %   if first character is a slash, that's it

    push_spot();
    while (what_line() != 1)
    {
        tree_parent();

        if (strlen(path))
            path = tree_node_name() + path;
    }
    pop_spot();

    return path;
}

define tree_show (root)
{
    %    Anchor the tree at 'root'

    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    variable tree_status_str = "  |  F1=Help  ";
    set_status_line (s_s+root+tree_status_str, 0);

    % Buffer status
    erase_buffer();
    bob();
    insert (root);

    () = tree_branch (root, s_s);
    tree_cwd = root;
    bob();

    set_buffer_modified_flag (0);

    %   when emerging from sub-directory
    %   highlight will stay on that sub-directory
    %   this is done in tree_parent_show
}

define tree_count_subs (ddir)
{
    %   ddir    must end with a slash
    %   tally of 1st level sub-dirs in ddir

    return lngth(listsubs (ddir));
}

define tree_branch (ddir, instr)
%----------------------------------------------------------
%   ddir    top directory of tree branch, ending with slash
%
%   instr   indentation string before each node.
%
%   directory is listed,
%   list is cleansed of everything not a directory,
%   sorted case-insensitive, displayed in a tree
%----------------------------------------------------------
{
    variable
        i = 0,
        A = listsubs(ddir),
        L = lngth(A);

    ifnot(L) return 0;

    %   paint the tree

    eol();
    newline();

    for (i=0;i<L;i++)
    {
        insert (instr);

        if (i == L-1) insert (cornr_p);
        else insert (vline_p);

        insert(A[i]);
        insert(slash);

        if (i != L-1) newline();
    }

    return L;
}

define tree_expand_one()
{
    set_readonly (0);
    EXIT_BLOCK set_readonly (1);

    variable
        s    = Null_String,
        sind = Null_String;

    %   variable 'sind' takes the string
    %   from bol to rightmost tree stroke

    if (bobp) return 0;

    push_spot();

    bol();
    push_mark();
    if(ffind(vline_p))
    {
        sind = bufsubstr();
        sind += vline_s;
    }

    else if(ffind(cornr_p))
    {
        sind = bufsubstr();
        sind += "   ";
    }

    else
    {
        pop_mark(0);
        pop_spot();
        return 0;
    }

    call ("next_char_cmd");
    del();
    insert ("-");

    s = tree_node_path ();

    %   prepare for insertion of branch
    variable tbr=0;
    tbr = tree_branch (s, sind);

    %   if no branch, delete empty line
    if(is_whiteline) delete_line();
    pop_spot();
    return tbr;
}

define tree_expand_enter()
%----------------------------------------------------------
%   if dir is logged (i.e. expanded one level)
%   pressing enter will list its files
%   otherwise it will expand it one level
%----------------------------------------------------------
{
    variable
        logged  = 0,
        col     = tree_point2(),
        s       = tree_node_path ();

    if (col==1) logged = 1;
    %   this takes care of top line
    %   top dir is always logged one level

    %   is dir logged?
    else if(bfind(vline_m) or bfind(cornr_m))
        logged = 1;

    push_spot();
    if(logged)
    {
        variable A = listdir(s);

        % are there any files?
        if (lngth(A))
        {
            delbuf(whatbuf);
            dls (s);
        }

        else flush("Directory is empty.");
    }

    else () = tree_expand_one();
    pop_spot();
}

define tree_collapse()
{
    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    if (what_line() == 1)
    {
        variable
            llstr  = line_as_string(),
            offset = string_match (llstr, "[^/]*[/]$", 1),
            subdir = substr (llstr, offset, -1);

        tree_parent_show();
        () = fsearch(p_s+subdir);
        return;
    }

    %   line is not 1

    () = tree_point2();

    go_left(3);
    if (looking_at(vline_p) or looking_at (cornr_p))
    {
        tree_parent();
        return;
    }
    else if(looking_at (cornr_m))
    {
        variable s = tree_node_path();

        ifnot(tree_count_subs (s))
        {
            () = tree_point2();
            call ("previous_char_cmd");
            call ("previous_char_cmd");
            del();
            insert ("+");
            () = tree_point2();
            return;
        }
    }

    eol();
    push_mark();
    %   mark must be at end of line for deleting

    %   move down to bottom of branch as displayed
    %   next sibling cannot do!

    tree_descend();
    eol();
    del_region();

    %   back to line where collapse command was issued

    bol();
    if(ffind(vline_m));
    else if(ffind(cornr_m));
    else
    {
        flush("Exception Error!");
        return;
    }

    %   either vline_m or cornr_m has been found

    go_right(1);
    del();
    insert ("+");

    () = tree_point2();
}

define tree_home()
{
    bob();
}

define tree_end()
{
    eob();
    () = bsearch(slash);
    () = tree_point2();
}

define tree_list_keys ()
{
    EXIT_BLOCK set_readonly (1);
    set_readonly (0);

    erase_buffer();
    bob();

    newline();
    dump_bindings ("tree_slimkmap");

    %   suppress lines containing self-insert characters
    bob ();
    while (fsearch ("tree_jump_to_initial")) delete_line ();

    readable_keys();

    bob();
    push_mark ();
    eob();
    bskip_chars(" \t\n");

    sort2 ();
    flush(" ");

    while (is_whiteline) del();
    %   this sort is not case sensitive

    eob();
    loop(8) newline();

    bob();
    newline();
    set_buffer_modified_flag (0);

    set_status_line("  Press any key to exit",0);

    variable m = "&Keybindings TreeView";
    dummy_menu(m);

    update (1);
    flush_input(getkey());

    menu_use_menu_bar (TreeView);
    tree_show (tree_cwd);
}

define tree_find()
{
    current_mini_cmd = _function_name;
    variable sstr=Null_String;
    sstr = read__mini ("Search for:", Null_String, Null_String, 0);
    ifnot (strlen (sstr)) return;
    LAST_SEARCH=sstr;
    () = fsearch(sstr);
}

define tree_find_again ()
{
    %   since the line and not the string is highlighted
    %   the string may occur more than once in same line
    %   and it is hard to notice
    eol();
    () = fsearch(LAST_SEARCH);
}

define tree_expand_9()
%----------------------------------------------------------
%   used mid-tree to expand only branch below current entry
%   without touching tree elsewhere, not used from top (root)
%   called only from tree_expand_all
%----------------------------------------------------------
{
    set_readonly (0);
    EXIT_BLOCK set_readonly (1);

    variable
        il   = 0,
        msg  = "seconds,  <ctrl><backslash> to stop",
        s    = Null_String,
        sind = Null_String,
        sz   = Null_String,
        tbr  = 0,
        tmpl = what_line(),
        tmpt = _time() - 1;

    %   already expanded?
    bol();
    if(ffind(vline_m) or ffind(cornr_m))
    {
        flush("Collapse directory first.");
        return;
    }

    %   we copy everything past current entry and delete the tail,
    %   when expansion is done, we append it back

    if(down(1))
    {
        ifnot(is_whiteline)
        {
            push_mark();
            () = dupmark();
            eob();
            sz = bufsubstr();
            del_region();
        }
        go_up(1);
        bol();
    }

    forever
    {
        if(fsearch(vline_p))
        {
            push_spot();
            push_mark();
            bol();
            sind = bufsubstr();
            sind += vline_s;
            pop_spot();
        }

        else if(fsearch(cornr_p))
        {
            push_spot();
            push_mark();
            bol();
            sind = bufsubstr();
            sind += "   ";
            pop_spot();
        }

        else if(bsearch(cornr_p))
        {
            goto_line(tmpl);
            continue;
        }

        else break;

        call ("next_char_cmd");
        del();
        insert ("-");

        _pop_n (_stkdepth);
        s = tree_node_path ();
        il = tree_branch (s, sind);
        if(is_whiteline) delete_line();
        tbr += il;

        flush(sprintf(" %d lines,  %d %s", tbr, _time() - tmpt, msg));
    }

    %   append the portion of tree previously deleted,
    %   in special cases it is just Null_String
    eob();
    insert(sz);
    goto_line(tmpl);
    flush(sprintf(" %d lines,  %d seconds.", tbr, _time() - tmpt));
}

define tree_expand_all()
{
    if(what_line > 1)
    {
        tree_expand_9();
        return;
    }

    set_readonly (0);
    EXIT_BLOCK set_readonly (1);

    variable
        il   = 0,
        msg  = "seconds,  <ctrl><backslash> to stop",
        s    = Null_String,
        sind = Null_String,
        tbr  = 0,
        tmpl = what_line(),
        tmpt = _time() - 1;

    forever
    {
        if(fsearch(vline_p))
        {
            push_spot();
            push_mark();
            bol();
            sind = bufsubstr();
            sind += vline_s;
            pop_spot();
        }

        else if(fsearch(cornr_p))
        {
            push_spot();
            push_mark();
            bol();
            sind = bufsubstr();
            sind += "   ";
            pop_spot();
        }

        else if(bsearch(cornr_p))
        {
            push_spot();
            push_mark();
            bol();
            sind = bufsubstr();
            sind += "   ";
            pop_spot();
        }

        else if(bsearch(vline_p))
        {
            push_spot();
            push_mark();
            bol();
            sind = bufsubstr();
            sind += vline_s;
            pop_spot();
        }

        else break;

        call ("next_char_cmd");
        del();
        insert ("-");

        _pop_n (_stkdepth);
        s = tree_node_path ();
        il = tree_branch (s, sind);
        if(is_whiteline) delete_line();
        tbr += il;

        flush(sprintf(" %d lines,  %d %s", tbr, _time() - tmpt, msg));
    }

    eob();
    flush(sprintf(" %d lines,  %d seconds.", -2+what_line, _time() - tmpt));
    goto_line(tmpl);
}


define tree_create_menu()
{
    %   The menu bar is created only once.
    %   When the create command is issued again,
    %   there is a check for the existence of the menu bar.
    %   To avoid garbage, delete before appending.

    menu_create_menu_bar (TreeView);
    menu_use_menu_bar (TreeView);

    menu_append_popup (TreeView, "&File");
    menu_append_popup (TreeView, "&Display");
    menu_append_popup (TreeView, "&Help");

    $1 = TreeView+".&File";

    menu_delete_items ($1);
    menu_append_item ($1, "&Open Listing..",    "tree_to_listing");
    menu_append_separator ($1);
    menu_append_item ($1, "&Save..",            "tree_save");
    menu_append_item ($1, "&Refresh",           "tree_refresh");
    menu_append_item ($1, "&Close",             "tree_quit");

    $1 = TreeView+".&Display";

    menu_delete_items ($1);
    menu_append_item ($1, "Upper &Directory",   "tree_upper_dir");
    menu_append_item ($1, "&Previous Sibling",  "tree_sibling_prev");
    menu_append_item ($1, "&Next Sibling",      "tree_sibling_next");
    menu_append_separator ($1);
    menu_append_item ($1, "&Top Directory",     "tree_top_root");
    menu_append_item ($1, "&User Directory",    "tree_user_root");
    menu_append_separator ($1);
    menu_append_item ($1, "&Expand One Level",  "tree_expand_one");
    menu_append_item ($1, "Expand &All",        "tree_expand_all");
    menu_append_item ($1, "&Collapse",          "tree_collapse");
    menu_append_separator ($1);
    menu_append_item ($1, "Current as &Root",   "tree_node_show");
    menu_append_item ($1, "&Show Fullpath",     "tree_node_show_path");

    $1 = TreeView+".&Help";

    menu_delete_items ($1);
    menu_append_item ($1, "&Quick Help",        "tree_display_help");
    menu_append_item ($1, "&List Keys",         "tree_list_keys");
}

define tree_keymap()
{
    reset_slim_keymap ();
    $1 = "tree_slimkmap";
    ifnot (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    undefinekey(Key_Del,$1);

    variable i;
    for (i=32; i<256; i++)
        definekey ("tree_jump_to_initial", char(i), $1);

    definekey ("tree_expand_all",       "*",        $1);
    definekey ("tree_find",             "^F",       $1);
    definekey ("tree_refresh",          "^R",       $1);
    definekey ("tree_save",             "^S",       $1);
    %definekey ("tree_test",             "^T",       $1);
    definekey ("tree_quit",             "^W",       $1);
    definekey ("tree_upper_dir",        Key_BS,     $1);
    definekey ("tree_move_down",        Key_Down,   $1);
    definekey ("tree_end",              Key_End,    $1);
    definekey ("tree_expand_enter",     Key_Enter,  $1);
    definekey ("tree_quit",             Key_Escape, $1);
    definekey ("tree_help",             Key_F1,     $1);
    definekey ("tree_menubar",          Key_F10,    $1);
    definekey ("tree_find_again",       Key_F3,     $1);
    definekey ("tree_to_listing",       Key_F4,     $1);
    definekey ("tree_home",             Key_Home,   $1);
    definekey ("tree_collapse",         Key_Left,   $1);
    definekey ("tree_pgdn",             Key_PgDn,   $1);
    definekey ("tree_pgup",             Key_PgUp,   $1);
    definekey ("tree_expand_one",       Key_Right,  $1);
    definekey ("tree_node_show",        Key_Space,  $1);
    definekey ("tree_sibling_next",     Key_Tab,    $1);
    definekey ("tree_move_up",          Key_Up,     $1);

    %   limit keys to avoid scrolling in keys list (help)
}

